# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""

The Graph datatypes. This brings together the scientific and framework methods
that are associated with the Graph datatypes.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Paula Sanz Leon <paula.sanz-leon@univ-amu.fr>

"""

from tvb.basic.traits import core, types_basic as basic
from tvb.basic.logger.builder import get_logger
from tvb.datatypes import arrays, time_series, connectivity

LOG = get_logger(__name__)


class Covariance(arrays.MappedArray):
    """Covariance datatype."""

    array_data = arrays.ComplexArray(file_storage=core.FILE_STORAGE_EXPAND)

    source = time_series.TimeSeries(
        label="Source time-series",
        doc="Links to the time-series on which NodeCovariance is applied.")

    __generate_table__ = True

    def configure(self):
        """After populating few fields, compute the rest of the fields"""
        # Do not call super, because that accesses data not-chunked
        self.nr_dimensions = len(self.read_data_shape())
        for i in range(self.nr_dimensions):
            setattr(self, 'length_%dd' % (i + 1), int(self.read_data_shape()[i]))

    def write_data_slice(self, partial_result):
        """
        Append chunk.
        """
        self.store_data_chunk('array_data', partial_result, grow_dimension=2, close_file=False)

    def _find_summary_info(self):
        summary = {"Graph type": self.__class__.__name__,
                   "Source": self.source.title}

        summary.update(self.get_info_about_array('array_data'))
        return summary


class CorrelationCoefficients(arrays.MappedArray):
    """Correlation coefficients datatype."""

    # Extreme values for pearson Correlation Coefficients
    PEARSON_MIN = -1
    PEARSON_MAX = 1

    array_data = arrays.FloatArray(file_storage=core.FILE_STORAGE_DEFAULT)

    source = time_series.TimeSeries(
        label="Source time-series",
        doc="Links to the time-series on which Correlation (coefficients) is applied.")

    labels_ordering = basic.List(
        label="Dimension Names",
        default=["Node", "Node", "State Variable", "Mode"],
        doc="""List of strings representing names of each data dimension""")

    __generate_table__ = True

    def configure(self):
        """After populating few fields, compute the rest of the fields"""
        # Do not call super, because that accesses data not-chunked
        self.nr_dimensions = len(self.read_data_shape())
        for i in range(self.nr_dimensions):
            setattr(self, 'length_%dd' % (i + 1), int(self.read_data_shape()[i]))

    def _find_summary_info(self):
        summary = {"Graph type": self.__class__.__name__,
                   "Source": self.source.title,
                   "Dimensions": self.labels_ordering}
        summary.update(self.get_info_about_array('array_data'))
        return summary

    def get_correlation_data(self, selected_state, selected_mode):
        matrix_to_display = self.array_data[:, :, int(selected_state), int(selected_mode)]
        return list(matrix_to_display.flat)


class ConnectivityMeasure(arrays.MappedArray):
    """Measurement of based on a connectivity."""

    connectivity = connectivity.Connectivity

    def _find_summary_info(self):
        summary = {"Graph type": self.__class__.__name__}
        # summary["Source"] = self.connectivity.title
        summary.update(self.get_info_about_array('array_data'))
        return summary

    __generate_table__ = True



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Mapped super-classes are defined here.

Important:

- Type - traited, possible mapped to db *col*
- MappedType - traited, mapped to db *table*


.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: marmaduke <mw@eml.cc>

"""

import six
import numpy
from scipy import sparse
from tvb.basic.logger.builder import get_logger
from tvb.basic.traits.util import get
from tvb.basic.traits.core import Type
from tvb.basic.traits.types_basic import DType


class MappedTypeLight(Type):
    """
    Light base class for all entities which are about to be mapped in storage.
    Current light implementation is to be used with the scientific-library stand-alone mode.
    """

    METADATA_EXCLUDE_PARAMS = ['id', 'LINKS', 'fk_datatype_group', 'disk_size',
                               'fk_from_operation', 'parent_operation', 'fk_parent_burst']

    ### Constants when retrieving meta-data about Array attributes on the current instance.
    METADATA_ARRAY_MAX = "Maximum"
    METADATA_ARRAY_MIN = "Minimum"
    METADATA_ARRAY_MEAN = "Mean"
    METADATA_ARRAY_VAR = "Variance"
    METADATA_ARRAY_MIN_NON_ZERO = "Min. non zero"
    METADATA_ARRAY_MAX_NON_ZERO = "Max. non zero"
    METADATA_ARRAY_MEAN_NON_ZERO = "Mean non zero"
    METADATA_ARRAY_VAR_NON_ZERO = "Var. non zero"
    METADATA_ARRAY_SHAPE = "Shape"
    _METADATA_ARRAY_SIZE = "Size"
    _METADATA_ARRAY_SIZE_NON_ZERO = "Size"

    DEFAULT_STORED_ARRAY_METADATA = [METADATA_ARRAY_MAX, METADATA_ARRAY_MIN, METADATA_ARRAY_MEAN,
                                     METADATA_ARRAY_VAR, METADATA_ARRAY_SHAPE]
    DEFAULT_WITH_ZERO_METADATA = [METADATA_ARRAY_MAX, METADATA_ARRAY_MIN, METADATA_ARRAY_MEAN,
                                  METADATA_ARRAY_VAR, METADATA_ARRAY_SHAPE, METADATA_ARRAY_MIN_NON_ZERO,
                                  METADATA_ARRAY_MEAN_NON_ZERO, METADATA_ARRAY_VAR_NON_ZERO]

    METADATA_FORMULAS = {METADATA_ARRAY_MAX: '$ARRAY$.max()',
                         METADATA_ARRAY_MIN: '$ARRAY$.min()',
                         METADATA_ARRAY_MEAN: '$ARRAY$.mean()',
                         METADATA_ARRAY_VAR: '$ARRAY$.var()',
                         METADATA_ARRAY_MAX_NON_ZERO: '$ARRAY$[$MASK$.nonzero()].max()',
                         METADATA_ARRAY_MIN_NON_ZERO: '$ARRAY$[$MASK$.nonzero()].min()',
                         METADATA_ARRAY_MEAN_NON_ZERO: '$ARRAY$[$MASK$.nonzero()].mean()',
                         METADATA_ARRAY_VAR_NON_ZERO: '$ARRAY$[$MASK$.nonzero()].var()',
                         METADATA_ARRAY_SHAPE: '$ARRAY$.shape()'}

    logger = get_logger(__name__)


    def __init__(self, **kwargs):
        super(MappedTypeLight, self).__init__(**kwargs)
        self._current_metadata = dict()


    @classmethod
    def from_file(cls, source_file="", instance=None):
        raise NotImplementedError("This DataType can not be used with load_default=True")


    def accepted_filters(self):
        """
        Just offer dummy functionality in library mode.
        """
        return {}

    @property
    def display_name(self):
        """
        To be implemented in each sub-class which is about to be displayed in UI,
        and return the text to appear.
        """
        return self.__class__.__name__


    def get_info_about_array(self, array_name, included_info=None, mask_array_name=None, key_suffix=''):
        """
        :return: dictionary {label: value} about an attribute of type mapped.Array
                 Generic information, like Max/Min/Mean/Var are to be retrieved for this array_attr
        """
        included_info = included_info or {}
        summary = self._get_summary_info(array_name, included_info, mask_array_name, key_suffix)
        ### Before return, prepare names for UI display.                
        result = dict()
        for key, value in six.iteritems(summary):
            result[array_name.capitalize().replace("_", " ") + " - " + key] = value
        return result


    def _get_summary_info(self, array_name, included_info, mask_array_name, key_suffix):
        """
        Get a summary from the metadata of the current array.
        :return: dictionary {label: value} about an attribute of type mapped.Array, with information like max/mean/etc
        """
        summary = dict()

        array_attr = getattr(self, array_name)
        if mask_array_name is not None:
            mask_attr = getattr(self, mask_array_name)
        else:
            mask_attr = array_attr

        if isinstance(array_attr, numpy.ndarray) and isinstance(mask_attr, numpy.ndarray):
            for key in included_info:
                if key in self.METADATA_FORMULAS:
                    summary[key + key_suffix] = eval(self.METADATA_FORMULAS[key].replace(
                        "$ARRAY$", "array_attr").replace("$MASK$", "mask_attr"))
                else:
                    self.logger.warning("Not supported meta-data will be ignored " + str(key))
        return summary


    def get_data_shape(self, data_name):
        """
        This method reads data-shape from the given data set
            ::param data_name: Name of the attribute from where to read size
            ::return: a shape tuple
        """
        array_data = getattr(self, data_name)
        if hasattr(array_data, 'shape'):
            return getattr(array_data, 'shape')
        self.logger.warning("Could not find 'shape' attribute on " + str(data_name) + " returning empty shape!!")
        return ()


    def get_data(self, data_name, data_slice=None):
        """
        Get sliced portion of an attribute of type numpy array.
        """
        array_data = getattr(self, data_name)
        return array_data[data_slice]


class Array(Type):
    """
    Traits type that wraps a NumPy NDArray.

    Initialization requires at least shape, and when not given, will be set to (), an empty, 0-dimension array.
    """

    wraps = numpy.ndarray
    dtype = DType()
    defaults = ((0,), {})
    data = None
    stored_metadata = [key for key in MappedTypeLight.DEFAULT_STORED_ARRAY_METADATA]
    logger = get_logger(__name__)


    @property
    def shape(self):
        """  
        Property SHAPE for the wrapped array.
        """
        return self.data.shape


    @property
    def array_path(self):
        """  
        Property PATH relative.
        """
        return self.trait.name


    def __get__(self, inst, cls):
        """
        When an attribute of class Array is retrieved on another class.
        :param inst: It is a MappedType instance
        :param cls: MappedType subclass. When 'inst' is None and only 'cls' is passed, we do not read from storage,
                    but return traited attribute.
        :return: value of type self.wraps
        :raise Exception: when read could not be executed, Or when used GET with incompatible attributes (e.g. chunks).
        """
        if inst is None:
            return self

        if self.trait.bound:
            return self._get_cached_data(inst)
        else:
            return self


    def __set__(self, inst, value):
        """
        This is called when an attribute of type Array is set on another class instance.
        :param inst: It is a MappedType instance
        :param value: expected to be of type self.wraps
        :raise Exception: When incompatible type of value is set
        """
        self._put_value_on_instance(inst, self.array_path)
        if isinstance(value, list):
            value = numpy.array(value)
        elif type(value) in (int, float):
            value = numpy.array([value])

        setattr(inst, '__' + self.trait.name, value)


    def _get_cached_data(self, inst):
        """
        Just read from instance since we don't have storage in library mode.
        """
        return get(inst, '__' + self.trait.name, None)


    def log_debug(self, owner=""):
        """
        Simple access to debugging info on a traited array, usage ::
            obj.trait["array_name"].log_debug(owner="obj")
            
        or ::
            self.trait["array_name"].log_debug(owner=self.__class__.__name__)
        """
        name = ".".join((owner, self.trait.name))
        sts = str(self.__class__)
        if self.trait.value is not None and self.trait.value.size != 0:
            shape = str(self.trait.value.shape)
            dtype = str(self.trait.value.dtype)
            tvb_dtype = str(self.trait.value.dtype)
            has_nan = str(numpy.isnan(self.trait.value).any())
            array_max = str(self.trait.value.max())
            array_min = str(self.trait.value.min())
            self.logger.debug("%s: %s shape: %s" % (sts, name, shape))
            self.logger.debug("%s: %s actual dtype: %s" % (sts, name, dtype))
            self.logger.debug("%s: %s tvb dtype: %s" % (sts, name, tvb_dtype))
            self.logger.debug("%s: %s has NaN: %s" % (sts, name, has_nan))
            self.logger.debug("%s: %s maximum: %s" % (sts, name, array_max))
            self.logger.debug("%s: %s minimum: %s" % (sts, name, array_min))
        else:
            self.logger.debug("%s: %s is Empty" % (sts, name))


class SparseMatrix(Array):
    """
    Map a big matrix.
    Will require storage in File Structure.
    """
    wraps = sparse.csc_matrix
    defaults = (((1, 1),), {'dtype': numpy.float64})
    logger = get_logger(__name__)


    def log_debug(self, owner=""):
        """
        Simple access to debugging info on a traited sparse matrix, usage ::
            obj.trait["sparse_matrix_name"].log_debug(owner="obj")
            
        or ::
            self.trait["sparse_matrix_name"].log_debug(owner=self.__class__.__name__)
        """
        name = ".".join((owner, self.trait.name))
        sts = str(self.__class__)
        if self.trait.value.size != 0:
            shape = str(self.trait.value.shape)
            sparse_format = str(self.trait.value.format)
            nnz = str(self.trait.value.nnz)
            dtype = str(self.trait.value.dtype)
            array_max = str(self.trait.value.data.max())
            array_min = str(self.trait.value.data.min())
            self.logger.debug("%s: %s shape: %s" % (sts, name, shape))
            self.logger.debug("%s: %s format: %s" % (sts, name, sparse_format))
            self.logger.debug("%s: %s number of non-zeros: %s" % (sts, name, nnz))
            self.logger.debug("%s: %s dtype: %s" % (sts, name, dtype))
            self.logger.debug("%s: %s maximum: %s" % (sts, name, array_max))
            self.logger.debug("%s: %s minimum: %s" % (sts, name, array_min))
        else:
            self.logger.debug("%s: %s is Empty" % (sts, name))



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
.. moduleauthor:: Calin Pavel <calin.pavel@codemart.ro>
.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
"""
import numpy
import json
from copy import deepcopy
from tvb.tests.library.base_testcase import BaseTestCase
import tvb.datatypes.equations as equations
import tvb.datatypes.arrays as arrays
import tvb.datatypes.time_series as time_series
from tvb.basic.traits.types_basic import MapAsJson
from tvb.basic.traits.types_mapped import MappedType
from tvb.basic.traits import types_basic as basic
from tvb.simulator.models import WilsonCowan, ReducedSetHindmarshRose


class TestTraits(BaseTestCase):
    """
    Test class for traits.core and traits.base
    """

    def test_default_attributes(self):
        """
        Test that default attributes are populated as they are described in Traits.
        """
        model_wc = WilsonCowan()
        assert type(model_wc.c_ii) == numpy.ndarray
        assert 1 == len(model_wc.c_ii)
        assert type(model_wc.tau_e) == numpy.ndarray
        assert 1 == len(model_wc.tau_e)
        assert isinstance(model_wc._tau_e, str)

    def test_modifying_attributes(self):
        """ 
        Test that when modifying an instance attributes, they are only visible on that instance.
        """
        eqn_t = equations.Gaussian()
        eqn_t.parameters["midpoint"] = 8.0

        eqn_x = equations.Gaussian()
        eqn_x.parameters["amp"] = -0.0625

        assert eqn_t.parameters is not None
        assert eqn_t.parameters['amp'] == 1.0
        assert eqn_t.parameters['midpoint'] == 8.0
        assert eqn_x.parameters is not None
        assert eqn_x.parameters['amp'] == -0.0625
        assert eqn_x.parameters['midpoint'] == 0.0

        a1 = arrays.FloatArray()
        a1.data = numpy.array([20.0, ])
        a2 = deepcopy(a1)
        a2.data = numpy.array([42.0, ])
        assert a1.data[0] != a2.data[0]

        model1 = ReducedSetHindmarshRose()
        model1.a = numpy.array([55.0, ])
        model1.number_of_modes = 10

        model2 = ReducedSetHindmarshRose()
        model2.a = numpy.array([42.0])
        model2.number_of_modes = 15
        assert model1.number_of_modes != model2.number_of_modes
        assert model1.a[0] != model2.a[0]

    def test_array_populated(self):
        """
        simple Array test
        """
        arr = arrays.MappedArray(array_data=numpy.array(range(10)))
        arr.configure()
        assert 10 == arr.array_data.shape[0]
        assert 1 == arr.nr_dimensions
        assert 10 == arr.length_1d

    def test_linked_attributes(self):
        """
        Test that tvb.core.trait.util produces something.
        """

        class Internal_Class(MappedType):
            """ Dummy persisted class"""
            x = 5
            z = basic.Integer()
            j = basic.JSONType()

            class In_Internal_Class(object):
                """Internal of Internal class"""
                t = numpy.array(range(10))

            @property
            def y(self):
                return self.x

        instance = Internal_Class()
        assert 5 == instance.y

        instance.j = {'dict_key': 10}
        assert isinstance(instance._j, str)
        assert isinstance(instance.j, dict)

    def test_none_complex_attribute(self):
        """
        Test that traited attributes are returned None, 
        when no value is assigned as default to them.
        """
        serie = time_series.TimeSeriesRegion()
        assert serie.connectivity is None

    def test_json_dumps_loads(self):
        """
        Tests class `MapAsJson.MapAsJsonEncoder` loads parameters correctly from JSON.
        """
        input_parameters = {'a': 1, 'b': 1.0, 'c': 'd'}
        test_dict = {'1': 1, 'a': {'1': 'b'}, '2': {'a': equations.Equation(parameters=input_parameters)}}
        json_string = json.dumps(test_dict, cls=MapAsJson.MapAsJsonEncoder)
        loaded_dict = json.loads(json_string, object_hook=MapAsJson.decode_map_as_json)
        eq_parameters = loaded_dict['2']['a']
        assert input_parameters == eq_parameters.parameters, "parameters not loaded properly from json"



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Calculate a ... on a .. datatype and return a ...

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy
#TODO: Currently built around the Simulator's 4D timeseries -- generalise...
import tvb.datatypes.time_series as time_series
import tvb.datatypes.graph as graph
import tvb.basic.traits.core as core
import tvb.basic.traits.util as util
from tvb.basic.logger.builder import get_logger

LOG = get_logger(__name__)




class NodeCovariance(core.Type):
    """
    Compute the temporal covariance of nodes in a TimeSeries dataType.
    A nodes x nodes matrix is returned for each (state-variable, mode).
    """

    time_series = time_series.TimeSeries(
        label="Time Series",
        required=True,
        doc="""The timeseries to which the NodeCovariance is to be applied.""")
    
    
    def evaluate(self):
        """
        Compute the temporal covariance between nodes in the time_series.
        """
        cls_attr_name = self.__class__.__name__ + ".time_series"
        self.time_series.trait["data"].log_debug(owner=cls_attr_name)
        
        data_shape = self.time_series.data.shape
        
        #(nodes, nodes, state-variables, modes)
        result_shape = (data_shape[2], data_shape[2], data_shape[1], data_shape[3])
        LOG.info("result shape will be: %s" % str(result_shape))
        
        result = numpy.zeros(result_shape)
        
        #One inter-node temporal covariance matrix for each state-var & mode.
        for mode in range(data_shape[3]):
            for var in range(data_shape[1]):
                data = self.time_series.data[:, var, :, mode]
                data = data - data.mean(axis=0)[numpy.newaxis, 0]
                result[:, :, var, mode] = numpy.cov(data.T)

        util.log_debug_array(LOG, result, "result")

        covariance = graph.Covariance(source=self.time_series,
                                      array_data=result,
                                      use_storage=False)
        return covariance
    
    
    def result_shape(self, input_shape):
        """
        Returns the shape of the main result of the NodeCovariance analysis.
        """
        result_shape = (input_shape[2], input_shape[2], input_shape[1], input_shape[3])
        return result_shape
    
    
    def result_size(self, input_shape):
        """
        Returns the storage size in Bytes of the NodeCovariance result.
        """
        result_size = numpy.prod(self.result_shape(input_shape)) * 8.0  # Bytes
        return result_size
    
    
    def extended_result_size(self, input_shape):
        """
        Returns the storage size in Bytes of the NodeCovariance extended result.
        That is, it includes storage of the evaluated PrincipleComponents
        attributes such as norm_source, component_time_series, etc.
        """
        extend_size = self.result_size(input_shape)  # Currently no derived attributes.
        return extend_size





# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
.. [AIMA_3rd_2010] Stuart Jonathan Russell, Peter Norvig, *Artificial 
intelligence: a modern approach*, 3rd Edition, Prentice Hall, 2010

This module was written for use in TVB as a way of determining a colouring 
scheme for the parcelated cortex. This represents a constraint satisfaction 
problem (CSP), see [AIMA_3rd_2010]_, chapter 6. The purpose is to return a 
colouring dictionary that satisfies the constraint of a region's colour not
being the same as any of its neighbours on the surface.

::
    import tvb.datatypes.surfaces as surfaces_module
    from tvb.simulator.region_boundaries import RegionBoundaries
    CORTEX = surfaces_module.Cortex()  
    CORTEX_BOUNDARIES = RegionBoundaries(CORTEX)
    region_colours = RegionColours(CORTEX_BOUNDARIES.region_neighbours)
    colouring = region_colours.back_track()


.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>

"""

import numpy

from tvb.simulator.common import get_logger
LOG = get_logger(__name__)


class RegionColours(object):
    """

    ''neighbours'':
        A dictionary specifying the neigbouring regions of each region.

    ''number_of_colours'':
        For a plane or a surface that is topologically spherical there is a
        theorem stating that only four colours are required to colour a surface
        without any neighbours having the same colour.
        NOTE: If you are considering the additional neighbour constraints
            imposed by the colourbar then it may be necessary to have more than
            four colour possibilities.

    """

    #TODO: Trait me...

    def __init__(self, neighbours, number_of_colours=4):
        """
        """
        self._neighbours = neighbours
        self._number_of_colours = number_of_colours
        self._regions = neighbours.keys()
        self._region_degree = [len(neighbours[r]) for r in self.regions]

        max_degree_index = numpy.array(self.region_degree).argmax()
        self._max_degree_region = self.regions[max_degree_index]

        domains = range(self.number_of_colours)
        self.possible_colours = dict()
        for region in self.neighbours:
            self.possible_colours[region] = list(domains) 

        self._colours = None
        self._region_colours = None


    @property
    def neighbours(self):
        """ A dictionary where specifying the neighbours of each region. """
        return self._neighbours


    @property
    def number_of_colours(self):
        """ The number of colours available for colouring the surface. """
        return self._number_of_colours


    @property
    def regions(self):
        """ A list of the regions comprising the surface. """
        return self._regions


    @property
    def region_degree(self):
        """ A list specifying the degree of each region. """
        return self._region_degree


    @property
    def max_degree_region(self):
        """ Maximum degree, variable selection heuristic. """
        return self._max_degree_region


#    @property
#    def region_colours(self):
#        """ A list specifying the colour of each region. """
#        return self.colours[]#Map region assignment to particular colours... 
#
#    
#    def _get_colours(self):
#        return self._colours
#    def _set_colours(self, colours):
#        if len(colours) == self.number_of_colours:
#            self._colours = colours
#        else:
#            LOG.error("A list of colours of length number_pf_colours must be provided")
#    property(colours, get=_get_colours set=_set_colours)


    def assigned(self, assignment):
        """ A list of the assigned regions """
        return [region for region in self.regions if region in assignment]


    def unassigned(self, assignment):
        """ A list of unassigned regions """
        return [region for region in self.regions if region not in assignment]


    def isconsistent(self, region, colour, assignment):
        """
        Check that using "colour" for "region" doesn't conflict with any of the
        colours already assigned to neighbouring regions.
        """
        for neighbour in self.neighbours[region]:
            if neighbour in self.assigned(assignment):
                if assignment[neighbour] == colour:
                    return False
        return True


    def mrv(self, assignment):
        """
        Minimum reamaining values, variable selection heuristic.
        """
        remaining = [len(self.possible_colours[r]) for r in self.unassigned(assignment)]
        return self.unassigned(assignment)[remaining.index(min(remaining))]

    #TODO: Might be good to add or replace with a least used weighting, for max diversity.
    def lcv(self, region):
        """
        Least constraining colour, value selection heuristic.
        """
        def neighbours_have(colour):
            weighted_count = 0.0
            for neighbour  in self.neighbours[region]:
                if colour in self.possible_colours[neighbour]:
                    weighted_count += 1.0 / len(self.possible_colours[neighbour])
            return weighted_count

        return sorted(self.possible_colours[region], key=neighbours_have)


    def forward_check(self, region, assignment):
        """
        Forward checking, inference step.
        """
        removed = dict()
        resolved = list()
        for neighbour in self.neighbours[region]:
            if ((neighbour not in assignment) and
                (assignment[region] in self.possible_colours[neighbour])):
                self.possible_colours[neighbour].remove(assignment[region])
                removed[neighbour] = assignment[region]
            if len(self.possible_colours[neighbour]) == 0:
                return (False, removed, resolved)
            elif len(self.possible_colours[neighbour]) == 1:
                assignment.update({neighbour: self.possible_colours[neighbour][0]})
                resolved.append(neighbour)
        return (True, removed, resolved)


    def iscomplete(self, assignment):
        """
        Checking for completion of assignment.
        """
        return (sorted(assignment.keys()) == sorted(self.regions))


    def back_track(self, assignment=None):
        """
        A backtracking algorithm for constraint satisfaction. The algorithm uses
        the degree heuristic for the initial variable selection, subsequent
        regions are selected via the minimum remaining values (MRV) heuristic.
        Colour selection for regions follows a least constraining value (LCV)
        heuristic. The backtracking search is augmented by a simple forward
        checking inference procedure. See, [AIMA_3rd_2010]_, Sec 6.3, Fig 6.5.

        """
        if assignment is None:
          assignment = dict()
            
        if self.iscomplete(assignment):
            LOG.debug("Solution found, returning assignment...")
            return assignment
        elif assignment == {}:
            region = self.max_degree_region
        else:
            region = self.mrv(assignment)
        #LOG.debug("For region %s" % (str(region)))
        #LOG.debug("Least constraining colour order is: %s" % (str(self.lcv(region))))
        for colour in self.lcv(region):
            if self.isconsistent(region, colour, assignment):
                #LOG.debug("Adding %s to %s" % (str(colour), str(region)))
                assignment.update({region: colour})
                self.possible_colours[region] = [colour]
                safe, removed, resolved = self.forward_check(region, assignment)
                if safe: #safe
                    result = self.back_track(assignment)
                    if result is not None:
                        return result
                #import pdb; pdb.set_trace()
                for neighbour in removed:
                    self.possible_colours[neighbour].append(removed[neighbour])
                for neighbour in resolved:
                    assignment.pop(neighbour)
        assignment.pop(region)
        #import pdb; pdb.set_trace()
        LOG.error("Failed to find an assignment...")
        return None




# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""
Created on Mar 20, 2013

.. moduleauthor:: Bogdan Neacsa <bogdan.neacsa@codemart.ro>
"""
from tvb.tests.library.base_testcase import BaseTestCase
from tvb.datatypes import arrays
from tvb.basic.traits.util import str_class_name
from tvb.basic.traits.core import FILE_STORAGE_DEFAULT


class TestCore(BaseTestCase):
    def test_traits_default(self):
        """
        Tests for default values upon creation of a FloatArray traited class.
        """
        array_dt = arrays.FloatArray()
        assert array_dt.trait.file_storage == FILE_STORAGE_DEFAULT
        assert array_dt.trait.order_number == 0
        assert array_dt.trait.required is True
        assert array_dt.trait.use_storage is True
        assert array_dt.trait.range_interval is None

    def test_traits_specific(self):
        """
        Tests for correct creation of a FloatArray traited class with initial values specified.
        """
        array_dt = arrays.FloatArray(file_storage="txt",
                                     order=6,
                                     required=False,
                                     use_storage=False,
                                     range=[1, 2, 3])
        assert array_dt.trait.file_storage == 'txt'
        assert array_dt.trait.order_number == 6
        assert array_dt.trait.required is False
        assert array_dt.trait.use_storage is False
        assert array_dt.trait.range_interval == [1, 2, 3]

    def test_str_class_name(self):
        """
        Tests method `tvb.basic.traits.util.str_class_name` works as expected with the FloatArray class
        """
        assert str_class_name(arrays.FloatArray) == 'tvb.datatypes.arrays.FloatArray'
        assert str_class_name(arrays.FloatArray, True) == 'FloatArray'
        assert str_class_name(1) == '1'



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and 
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#

"""
Generate Dictionary required by the Framework to generate UI from it.
Returned dictionary will be generated from  traited definition of attributes.


.. moduleauthor:: Lia Domide <lia.domide@codemart.ro>
.. moduleauthor:: Stuart Knock <stuart.knock@gmail.com>
.. moduleauthor:: marmaduke <duke@eml.cc>

"""

import numpy
import json
from tvb.basic.logger.builder import get_logger
from tvb.basic.traits.util import get, str_class_name, multiline_math_directives_to_matjax
from tvb.basic.traits.core import KWARG_AVOID_SUBCLASSES, TYPE_REGISTER, KWARG_FILTERS_UI, KWARG_FILTERS_BACKEND

LOG = get_logger(__name__)

INTERFACE_ATTRIBUTES_ONLY = "attributes-only"
INTERFACE_ATTRIBUTES = "attributes"



class TraitedInterfaceGenerator(object):
    """
    Bases class for interface reading and dumping. As a data descriptor, when 
    it is an attribute of the class it will compute a dictionary and return it.
    """


    def __get__(self, inst, ownr):

        obj = inst if inst else ownr
        if not obj.trait.bound:
            return {}

        intr = {}
        self.__fill_basics(obj, intr)
        self.__fill_ranges(obj, intr)
        self.__fill_noiseconfig(obj, intr)
        self.__fill_filters(obj, intr)

        if obj.trait.wraps is not None:
            self.__fill_wrapped_type(obj, intr)

            if intr['type'] == 'dict' and isinstance(intr['default'], dict):
                intr['attributes'], intr['elementType'] = self.__prepare_dictionary(intr['default'])
                if len(intr['attributes']) == 0:
                    ## Dictionary without any sub-parameter
                    return {}

        mro_type_names = [i.__name__ for i in ownr.mro()]

        if 'Array' in mro_type_names:
            self.__fill_array(obj, inst, intr)
        ##### TYPE & subclasses specifics ######################################
        elif ('Type' in mro_type_names
              and (obj.__module__ != 'tvb.basic.traits.types_basic' or 'Range' in mro_type_names)
              or 'Enumerate' in mro_type_names):

            self.__fill_entity_attributes(obj, intr)

            if obj.trait.bound == INTERFACE_ATTRIBUTES_ONLY:
                # We need to do this, to avoid infinite loop on attributes 
                # of class Type with no subclasses
                return intr

            if obj.trait.select_multiple:
                intr['type'] = 'selectMultiple'
            else:
                intr['type'] = 'select'

            if 'MappedType' in mro_type_names:
                self.__fill_mapped_type(ownr, intr)
            else:
            ##### TYPE (not MAPPED_TYPE) again ####################################
                intr['attributes'] = []
                # Build options list
                intr['options'] = []
                if 'Enumerate' in obj.__class__.__name__:
                    self.__fill_enumerate(obj, intr)
                    return intr
                else:
                    self.__handle_nonmapped_subtypes(ownr, obj, intr)

            self.__correct_defaults(intr)

        return intr


    @staticmethod
    def __fill_basics(obj, intr):
        label = get(obj.trait.inits.kwd, 'label', obj.trait.name)
        if not label:
            label = obj.trait.name
        intr.update({
            'default': (obj.value or obj.trait.value) if hasattr(obj, 'value') else obj.trait.value,
            'description': get(obj.trait.inits.kwd, 'doc'),
            'label': label.capitalize(),
            'name': obj.trait.name,
            'locked': obj.trait.inits.kwd.get('locked', False),
            'required': obj.trait.inits.kwd.get('required', True)
        })


    @staticmethod
    def __fill_ranges(obj, intr):
        range_value = obj.trait.inits.kwd.get('range', False)
        if range_value:
            intr['minValue'] = range_value.lo
            intr['maxValue'] = range_value.hi
            if range_value.step is not None:
                intr['stepValue'] = range_value.step
            else:
                LOG.debug("Missing Range.step field for attribute %s, we will consider a default." % obj.trait.name)
                intr['stepValue'] = (range_value.hi - range_value.hi) / 10


    @staticmethod
    def __fill_noiseconfig(obj, intr):
        noise_configurable = obj.trait.inits.kwd.get('configurable_noise', None)
        if noise_configurable is not None:
            intr['configurableNoise'] = noise_configurable


    @staticmethod
    def __fill_filters(obj, intr):
        if KWARG_FILTERS_UI in obj.trait.inits.kwd:
            intr[KWARG_FILTERS_UI] = json.dumps([ui_filter.to_dict() for ui_filter in
                                                 obj.trait.inits.kwd[KWARG_FILTERS_UI]])
        if KWARG_FILTERS_BACKEND in obj.trait.inits.kwd:
            intr["conditions"] = obj.trait.inits.kwd[KWARG_FILTERS_BACKEND]


    @staticmethod
    def __fill_wrapped_type(obj, intr):
        if isinstance(obj.trait.wraps, tuple):
            intr['type'] = obj.trait.wraps[0].__name__
        else:
            intr['type'] = obj.trait.wraps.__name__


    @staticmethod
    def __fill_array(obj, inst, intr):
        intr['type'] = 'array'
        intr['elementType'] = str(inst.dtype)
        if isinstance(obj.trait.value, numpy.ndarray):
            # Make sure arrays are displayed in a compatible form: [1, 2, 3]
            intr['default'] = str(obj.trait.value.tolist())


    @staticmethod
    def __fill_entity_attributes(obj, intr):
        # Populate Attributes for current entity
        attrs = sorted(obj.trait.values(), key=lambda entity: entity.trait.order_number)
        attrs = [val.interface for val in attrs if val.trait.order_number >= 0]
        attrs = [attr for attr in attrs if attr is not None and len(attr) > 0]
        intr['attributes'] = attrs


    @staticmethod
    def __fill_mapped_type(ownr, intr):
        intr['datatype'] = True
        #### For simple DataTypes, cut options and attributes
        intr['options'] = []
        if not ownr._ui_complex_datatype:
            intr['attributes'] = []
            ownr_class = ownr.__class__
        else:
            ownr_class = ownr._ui_complex_datatype
        if 'MetaType' in ownr_class.__name__:
            ownr_class = ownr().__class__
        intr['type'] = ownr_class.__module__ + '.' + ownr_class.__name__


    @staticmethod
    def __fill_enumerate(obj, intr):
        for val in obj.trait.options:
            intr['options'].append({'name': val,
                                    'value': val})
        intr['default'] = obj.trait.value


    def __prepare_dictionary(self, dictionary):
        """
        From base.Dict -> default [isinstance(dict)], prepare an interface specific tree.
        """
        result = []
        element_type = None
        for key in dictionary:
            entry = {}
            value = dictionary[key]
            entry['label'] = key
            entry['name'] = key
            if type(value) == dict:
                entry['attributes'], entry['elementType'] = self.__prepare_dictionary(value)
                value = ''
            entry['default'] = str(value)

            if hasattr(value, 'tolist') or 'Array' in [i.__name__ for i in type(value).mro()]:
                entry['type'] = 'array'
                if not hasattr(value, 'tolist'):
                    entry['default'] = str(value.trait.value)
            else:
                entry['type'] = type(value).__name__

            element_type = entry['type']
            result.append(entry)
        return result, element_type


    @staticmethod
    def __handle_nonmapped_subtypes(ownr, obj, intr):
        """ Populate options for each subtype. This fills in models etc"""
        for opt in TYPE_REGISTER.subclasses(ownr, KWARG_AVOID_SUBCLASSES in obj.trait.inits.kwd):
            if hasattr(obj, 'value') and obj.value is not None and isinstance(obj.value, opt):
                ## fill option currently selected with attributes from instance
                opt = obj.value
                opt_class = opt.__class__
            else:
                opt_class = opt
            opt.trait.bound = INTERFACE_ATTRIBUTES_ONLY

            description = multiline_math_directives_to_matjax(opt_class.__doc__)

            intr['options'].append({'name': get(opt, '_ui_name', opt_class.__name__),
                                    'value': str_class_name(opt_class, short_form=True),
                                    'class': str_class_name(opt_class, short_form=False),
                                    'description': description,
                                    'attributes': opt.interface['attributes']})


    @staticmethod
    def __correct_defaults(intr):
        if intr['default'] is not None:
            intr['default'] = intr['default'].__class__.__name__
            if intr['default'] == 'RandomState':
                intr['default'] = 'RandomStream'
        else:
            intr['default'] = None


    def __set__(self, inst, val):
        """
        Given a hierarchical dictionary of the kind generated by __get__, with the 
        chosen options, we should be able to fully instantiate a class.
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-
#
#
#  TheVirtualBrain-Scientific Package. This package holds all simulators, and
# analysers necessary to run brain-simulations. You can use it stand alone or
# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)

"""
A collection of neuronal dynamics models.

Specific models inherit from the abstract class Model, which in turn inherits
from the class Trait from the tvb.basic.traits module.

.. moduleauthor:: Stuart A. Knock <Stuart@tvb.invalid>
.. moduleauthor:: Paula Sanz Leon <Paula@tvb.invalid>
.. moduleauthor:: Gaurav Malhotra <Gaurav@tvb.invalid>
.. moduleauthor:: Marmaduke Woodman <marmaduke.woodman@univ-amu.fr>

"""

from .base import Model
from .wilson_cowan import WilsonCowan
from .stefanescu_jirsa import ReducedSetFitzHughNagumo, ReducedSetHindmarshRose
from .jansen_rit import JansenRit, ZetterbergJansen
from .oscillator import Generic2dOscillator, Kuramoto
from .larter_breakspear import LarterBreakspear
from .wong_wang import ReducedWongWang
from .wong_wang_exc_io_inh_i import ReducedWongWangExcIOInhI
from .linear import Linear
from .hopfield import Hopfield
from .epileptor import Epileptor, Epileptor2D
from .JCepileptor import JC_Epileptor
from .epileptorcodim3 import EpileptorCodim3, EpileptorCodim3SlowMod
from .Zerlaut import Zerlaut_adaptation_first_order, Zerlaut_adaptation_second_order



# in conjunction with TheVirtualBrain-Framework Package. See content of the
# documentation-folder for more details. See also http://www.thevirtualbrain.org
#
# (c) 2012-2017, Baycrest Centre for Geriatric Care ("Baycrest") and others
#
# This program is free software: you can redistribute it and/or modify it under the
# terms of the GNU General Public License as published by the Free Software Foundation,
# either version 3 of the License, or (at your option) any later version.
# This program is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
# PARTICULAR PURPOSE.  See the GNU General Public License for more details.
# You should have received a copy of the GNU General Public License along with this
# program.  If not, see <http://www.gnu.org/licenses/>.
#
#
#   CITATION:
# When using The Virtual Brain for scientific publications, please cite it as follows:
#
#   Paula Sanz Leon, Stuart A. Knock, M. Marmaduke Woodman, Lia Domide,
#   Jochen Mersmann, Anthony R. McIntosh, Viktor Jirsa (2013)
#       The Virtual Brain: a simulator of primate brain network dynamics.
#   Frontiers in Neuroinformatics (7:10. doi: 10.3389/fninf.2013.00010)
#
#
"""

Floating-point range generator (IEEE-754-proof)
Original: https://gist.github.com/diogobaeder/1239977


.. moduleauthor:: Paula Sanz Leon <paula.sanz-leon@unvi-amu.fr>
"""

from tvb.tests.library.base_testcase import BaseTestCase
from tvb.basic.traits.types_basic import Range


class TestRange(BaseTestCase):
    def test_generates_range_with_only_stop_provided(self):
        floats = list(Range(hi=3.0, step=1.0))
        assert floats == [0.0, 1.0, 2.0]

    def test_generates_range_with_start_and_stop(self):
        floats = list(Range(lo=1.0, hi=3.0, step=1.0))
        assert floats == [1.0, 2.0]

    def test_generates_range_with_start_stop_and_step_smaller_than_one(self):
        floats = list(Range(lo=0.0, hi=0.5, step=0.1))
        assert floats == [0.0, 0.1, 0.2, 0.3, 0.4]

    def test_generates_range_with_start_stop_and_periodic_repeating_step(self):
        floats = list(Range(lo=0.0, hi=1.0, step=1. / 3))
        assert floats == [0.0, 0.333333333333, 0.666666666666, 0.999999999999]

    def test_generates_range_with_start_and_end_included(self):
        floats = list(Range(lo=0.0, hi=3.0, step=1.0, mode=Range.MODE_INCLUDE_BOTH))
        assert floats == [0.0, 1.0, 2.0, 3.0]

    def test_generates_range_with_start_and_end_excluded(self):
        floats = list(Range(lo=0.0, hi=3.0, step=1.0, mode=Range.MODE_EXCLUDE_BOTH))
        assert floats == [1.0, 2.0]

    def test_generates_range_with_only_end_included(self):
        floats = list(Range(lo=0.0, hi=3.0, step=1.0, mode=Range.MODE_INCLUDE_END))
        assert floats == [1.0, 2.0, 3.0]

    def test_generates_range_with_negative_end(self):
        floats = list(Range(lo=1.0, hi=-3.0, step=1.0))
        assert floats == [1.0, 0.0, -1.0, -2.0]

    def test_generates_range_with_negative_end_including_both(self):
        floats = list(Range(lo=1.0, hi=-3.0, step=1.0, mode=Range.MODE_INCLUDE_BOTH))
        assert floats == [1.0, 0.0, -1.0, -2.0, -3.0]

    def test_generates_range_with_negative_end_but_excluding_start(self):
        floats = list(Range(lo=1.0, hi=-3.0, step=1.0, mode=Range.MODE_INCLUDE_END))
        assert floats == [0.0, -1.0, -2.0, -3.0]

    def test_generates_range_with_challenging_float_point_arithmetics(self):
        floats = list(Range(lo=0.0, hi=2.2, step=0.7))
        assert floats == [0.0, 0.7, 1.4, 2.1]
